/*
http://codeandlife.com/2012/06/18/usb-hid-keyboard-with-v-usb/
http://www.freebsddiary.org/APC/usb_hid_usages.php
http://www.microchip.com/forums/m618147.aspx
*/

/*
	LIBUSB_SUCCESS = 0,
	LIBUSB_ERROR_IO = -1,
	LIBUSB_ERROR_INVALID_PARAM = -2,
	LIBUSB_ERROR_ACCESS = -3,
	LIBUSB_ERROR_NO_DEVICE = -4,
	LIBUSB_ERROR_NOT_FOUND = -5,
	LIBUSB_ERROR_BUSY = -6,
	LIBUSB_ERROR_TIMEOUT = -7,
	LIBUSB_ERROR_OVERFLOW = -8,
	LIBUSB_ERROR_PIPE = -9,
	LIBUSB_ERROR_INTERRUPTED = -10,
	LIBUSB_ERROR_NO_MEM = -11,
	LIBUSB_ERROR_NOT_SUPPORTED = -12,
	LIBUSB_ERROR_OTHER = -99,
*/

#include "usbportability.h"
#include "usbdrv.h"
#include "oddebug.h"
#include <util/delay.h>
#include <avr/eeprom.h>
#define nop() asm volatile("nop\n\t"::);

#define LED_ON()  do { PORTB |=  0b00000010; } while (0)
#define LED_OFF() do { PORTB &= ~0b00000010; } while (0)
		 

// ************************
// *** USB HID ROUTINES ***
// ************************

// From Frank Zhao's USB Business Card project
// http://www.frank-zhao.com/cache/usbbusinesscard_details.php
const PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {

    0x05, 0x01,
    0x09, 0x80,
    0xa1, 0x01,
    0x85, 0x01,
    0x19, 0x81,
    0x29, 0x83,
    0x15, 0x00,
    0x25, 0x01,
    0x95, 0x03,
    0x75, 0x01,

    0x81, 0x02,
    0x95, 0x01,
    0x75, 0x05,
    0x81, 0x01,
    0xc0,

    0x05, 0xc,
    0x09, 0x01,
    0xa1, 0x01,
    0x85, 0x02,
    0x15, 0x00,

    0x25, 0x01,
    0x09, 0xe9,
    0x09, 0xea,
    0x09, 0xe2,
    0x09, 0xcd,
    0x19, 0xb5,
    0x29, 0xb8,
    0x75, 0x01,
    0x95, 0x08,
    0x81, 0x02,
		
    0x0a, 0x8a, 0x01,
    0x0a, 0x21, 0x02,
		0x0a, 0x2a, 0x02,
		0x1a, 0x23, 0x02,
		0x2a, 0x27, 0x02,
		0x81, 0x02,
		
		0x0a, 0x83, 0x01,
		0x0a, 0x96, 0x01,
		0x0a, 0x92, 0x01,
		0x0a, 0x9e, 0x01,
		0x0a, 0x94, 0x01,
		0x0a, 0x06, 0x02,
		0x09, 0xb2, 
		0x09, 0xb4,
		0x81, 0x02, 

    0xc0                           // END_COLLECTION
};

/*
typedef struct {
  uint8_t modifier;
  uint8_t reserved;
  uint8_t keycode[6];
} keyboard_report_t;
*/

typedef struct {
  uint8_t keycode[4];
} keyboard_report_t;



static keyboard_report_t keyboard_report; // sent to PC
volatile static uchar LED_state = 0xff; // received from PC
static uchar idleRate; // repeat rate for keyboards

usbMsgLen_t usbFunctionSetup(uchar data[8]) {
	usbRequest_t *rq = (void *)data;

  if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {
    switch(rq->bRequest) {
    case USBRQ_HID_GET_REPORT: // send "no keys pressed" if asked here
      // wValue: ReportType (highbyte), ReportID (lowbyte)
			usbMsgPtr = (void *)&keyboard_report; // we only have this one
			keyboard_report.keycode[0] = 0;
			keyboard_report.keycode[1] = 0;
			return sizeof(keyboard_report);
    case USBRQ_HID_SET_REPORT: // if wLength == 1, should be LED state
      return (rq->wLength.word == 1) ? USB_NO_MSG : 0;
    case USBRQ_HID_GET_IDLE: // send idle rate to PC as required by spec
			usbMsgPtr = &idleRate;
			return 1;
    case USBRQ_HID_SET_IDLE: // save idle rate as required by spec
			idleRate = rq->wValue.bytes[1];
			return 0;
    }
  }
	return 0; // by default don't return any data
}

#define NUM_LOCK 1
#define CAPS_LOCK 2
#define SCROLL_LOCK 4

usbMsgLen_t usbFunctionWrite(uint8_t * data, uchar len) {
  if (data[0] == LED_state)
		return 1;
	else
		LED_state = data[0];
	
	// LED state changed
	if(LED_state & CAPS_LOCK)
		PORTB |= 1 << PB1; // LED on
  else
    PORTB &= ~(1 << PB1); // LED off
	
  return 1; // Data read, not expecting more
}

// Now only supports letters 'a' to 'z' and 0 (NULL) to clear buttons
void buildReport(uchar send_key) {
  //keyboard_report.modifier = 0;
	
  if(send_key >= 'a' && send_key <= 'z') {
    keyboard_report.keycode[0] = 0x01;
    keyboard_report.keycode[1] = 0x02;
	}	else {
    keyboard_report.keycode[0] = 0x01;
    keyboard_report.keycode[1] = 0x00;
	}
		
}

#define STATE_WAIT 0
#define STATE_SEND_KEY 1
#define STATE_RELEASE_KEY 2
#define STATE_WAKEUP 3
#define STATE_STOP 4

/* Add new function usbFunctionRead
uchar usbFunctionRead(uchar *data, uchar len) {
	return 1; // equals the amount of bytes written
}
*/


int main() {
  uchar i, state = STATE_WAIT;
	uchar wakeupEvent=0;

	unsigned int button_push_counter = 0;
	uchar event_counter = 0;
	
		
  DDRB  = _BV(PB1); // PB1 as output
  PORTB = _BV(PB0); // PB0 is input with internal pullup resistor activated
  
	for(i=0; i<sizeof(keyboard_report); i++) // clear report initially
		((uchar *)&keyboard_report)[i] = 0;
    
	//wdt_enable(WDTO_1S); // enable 1s watchdog timer

	usbInit();
  
	i=10;
	while(--i){ _delay_ms(100);	}

	usbDeviceDisconnect(); // enforce re-enumeration
  for(i = 0; i<250; i++) { // wait 500 ms
		// wdt_reset(); // keep the watchdog happy
		_delay_ms(2);
  }
	usbDeviceConnect();
  
	TCCR0B |= (1 << CS01); // timer 0 at clk/8 will generate randomness
    
	sei(); // Enable interrupts after re-enumeration

	// volatile static unsigned int IntCnt;  
  while(1) {

		/*
		if (state==STATE_WAIT) {
			static int cnt;
			cnt++;
			if (cnt<200) LED_ON(); else LED_OFF();
			if (cnt>400) cnt=0;
		}
		
		if (state==STATE_SEND_KEY || state==STATE_RELEASE_KEY) {
			waitCnt++;
			if (waitCnt > 50) {
				state=STATE_WAKEUP;
				waitCnt=0;
			}
		} else {
			waitCnt=0;
		}
		*/
		
		/*
		{
			volatile static unsigned int prev;
			volatile static unsigned int noIntCnt=0;

			if (prev==IntCnt && noIntCnt<0xfff) { noIntCnt++; } else { noIntCnt=0; }
			if (noIntCnt == 0xfff) { PORTB |=  _BV(PB1); } // ڥ
			else                   { PORTB &= ~_BV(PB1); } // ̾
			prev=IntCnt;
		}
		*/
		
		// wdt_reset(); // keep the watchdog happy
		usbPoll();
		
		// button pressed (PB0 at ground voltage)
    if (!(PINB & (1<<PB0))) { 
			if (button_push_counter < 65535) { button_push_counter++; }
			_delay_ms(1);
		} else {
			// ꡼
			if (event_counter==255 && button_push_counter>0) {
				if (button_push_counter > 1000) {
					// Ĺ
					state = STATE_SEND_KEY;
				} else {
					wakeupEvent=1;
				// û
				}
				event_counter=0;
			}
			button_push_counter=0;
		}			
		if (event_counter < 255) { event_counter++; } // increase release counter


		if (wakeupEvent) {
			cli();
			uint8_t ddr = DDRB, ddr_init = DDRB; // read direction
			ddr |= USBMASK; // set output for D+ and D-
			USBOUT |= (1<< USBMINUS); // set J state in port
			USBDDR = ddr; // acquire bus
			USBOUT ^= USBMASK;				// set k state
			_delay_ms(10);			// wait
			USBOUT ^= USBMASK;			// set idle
			USBDDR = ddr_init;			// revert ddr
			USBOUT &= ~( 1 << USBMINUS );				// set port without pullup ie D+,D- = 0
			sei();
			wakeupEvent=0;
		}
		
		// characters are sent when messageState == STATE_SEND and after receiving
		// the initial LED state from PC (good way to wait until device is recognized)
		// if(usbInterruptIsReady()) IntCnt++;
		if(usbInterruptIsReady() && state != STATE_WAIT) {
			switch(state) {
			case STATE_SEND_KEY:
				buildReport('x');
				state = STATE_RELEASE_KEY; // release next
				break;
			case STATE_RELEASE_KEY:
				buildReport(0);
				state = STATE_WAIT; // go back to waiting
				break;
			default:
				state = STATE_WAIT; // should not happen
			}
			// start sending
			usbSetInterrupt((void *)&keyboard_report, sizeof(keyboard_report));
		}
  }
	return 0;
}

